#!/usr/bin/python
# -*- coding: utf-8 -*-

#############################################################################
# Version: 1 (Jun, 2014)                                                    #
# Author: Roberta Montefusco (rrt.montefusco@gmail.com), University of Pisa #
#                                                                           # 
# Evaluates matching relations                                              #
#############################################################################

"""
NAME:
  evalRels.py - evaluates matching relations.
  
SYNOPSIS:
  ./evalRels.py -g gold_file -s system_file [-o output_file] [OPTIONS]

DESCRIPTION:
  Returns the evalutation of system output with respect to a gold standard.
  Both files should be in UTF-8 encoded CoNLL-07 tabular format.
  Punctuation is not scored.
  
OPTIONS:
  -d	Returns the detailed measures.
	
OTHER OPTIONS:
  -h	Output this manual and terminates.
  
#############################################################################
# Version: 1 (Jun, 2014)                                                    #
# Author: Roberta Montefusco (rrt.montefusco@gmail.com), University of Pisa #
#############################################################################
"""


import getopt
import sys
import codecs
import string
import unicodedata



# nome del file di errore, file e messaggi di errore
# errorFileName = "EVALerrors"

# ferr = ""



# stampa mess sul file ferr e stampa il manuale se h = True
def exit(mess, h):
	# global ferr
	# ferr.write(mess)
	sys.stderr.write("\n" + mess)
	if h:
		# ferr.write(__doc__)
		sys.stderr.write(__doc__)
	sys.stderr.write("\n")
	# ferr.close()
	sys.exit()
	
	

# nome del file di errore
# errorFileNameG = "MatchErrorsGOLD"
# errorFileNameS = "MatchErrorsSYS"

	

# visualizza le misure di tutte le dipendenze, separatamente e aggiorna il dizionario inserendo per ogni dipendenza,
# una tripla (FG, FS, C1, C2) : numero di occorrenze trovate nel gold e nel sys, corrette come DEPREL (Acc) e come
# DEPREL + ATTACH (LAS: LAS + Acc).

diff_measures = False

all_measures = {}		

exclude_dep = ['ROOT', 'appos', 'attr', 'aux', 'auxpass', 'det', 'cc', 'conj', 'conj_e', 'conj_negcc', 'conj_o', 'csubj', 'predet', 'prt', 'expl', 'conj', 'dep', 'mwe', 'neg', 'num', 'number', 'parataxis', 'poss', 'pobj', 'preconj', 'prep', 'pcomp', 'quantmod']		
				
				
# misure da calcolare
class Measures:
	def __init__(self):
		self.gLAS = self.gUAS = self.gAcc = self.gtot = 0
		self.sLAS = self.sUAS = self.sAcc = self.stot = 0
		
	
	def update(self, measures):
		self.gLAS += int(measures.gLAS)
		self.gUAS += int(measures.gUAS)
		self.gAcc += int(measures.gAcc)
		self.gtot += int(measures.gtot)
		self.sLAS += int(measures.sLAS)
		self.sUAS += int(measures.sUAS)
		self.sAcc += int(measures.sAcc)
		self.stot += int(measures.stot)
		
		
		
# riga di una frase
class Row:
	def __init__(self, line, numline):
		index1 = line.find("(")
		index2 = line.find(" ") - 1
		sep1 = line.rfind("-", 0 , index2)
		sep2 = line.rfind("-", index2)
		copyheadindex = line.find("'", sep1, index2)	
		copybodyindex = line.find("'", sep2)
		
		self.dep = line[:index1]
		if self.dep.lower() == "root":
			self.dep = "ROOT"
			
		self.head = line[index1 + 1 : sep1]
		if copyheadindex != -1:
			if (line[sep1 + 1 : copyheadindex]).isdigit():
				self.nhead = int(line[sep1 + 1 : copyheadindex])
			else:
				exit("Error in format file at line %d. For info, see manual:" % numline, True)
			self.copyhead = line[copyheadindex : index2]
		else:
			if (line[sep1 + 1 : index2]).isdigit():
				self.nhead = int(line[sep1 + 1 : index2])
			else:
				exit("Error in format file at line %d. For info, see manual:" % numline, True)
			self.copyhead = ""
		
		self.body = line[index2 + 2 : sep2]
		if copybodyindex != -1:
			if (line[sep2 + 1 : copybodyindex]).isdigit():
				self.nbody = int(line[sep2 + 1 : copybodyindex])
			else:
				exit("Error in format file at line %d. For info, see manual:" % numline, True)
			self.copybody = line[copybodyindex : -1]
		else:
			if (line[sep2 + 1 : -1]).isdigit():
				self.nbody = int(line[sep2 + 1 : -1])
			else:
				exit("Error in format file at line %d. For info, see manual:" % numline, True)
			self.copybody = ""	
			
		self.info = ""
	
	
	# verifica se le dipendenze delle due righe matchano
	def matchDEP(self, other):
		return self.dep.lower() == other.dep.lower()
		
	
	# verifica se i token head delle righe matchano
	def matchHEAD(self, other):
		return self.head.lower() == other.head.lower() and self.nhead == other.nhead
		
		
	# verifica se i token body delle righe matchano
	def matchBODY(self, other):
		return self.body.lower() == other.body.lower() and self.nbody == other.nbody
					
	
	# porta in stringa una riga
	def __str__(self):
		return ("%s(%s-%d%s, %s-%d%s)" % (self.dep, self.head, self.nhead, self.copyhead, self.body, self.nbody, self.copybody)).encode('utf-8')
		
		

# frase		
class Sentence:	
	def __init__(self, rows):
		self.rows = rows
		
	
	# porta in stringa una frase	
	def __str__(self):
		s = ""
		for x in self.rows:
			s += str(x) + "\n"
		return s



# OUTPUT BASE:
# dati i valori passati come parametro, calcola la precision, recall e f1 e restituisce una stringa con il messaggio formattato
def getStringResult(measures):
	if measures.stot:
		LAS = float(measures.sLAS) / (measures.stot)
		UAS = float(measures.sUAS) / (measures.stot)
		Acc = float(measures.sAcc) / (measures.stot)
		Prec = (LAS, UAS)
		output = ("\n  Precision (LAS): %d / %d * 100 = %.2f%%\n  Precision (UAS): %d / %d * 100 = %.2f%%\n  System accuracy: " % \
		(measures.sLAS, measures.stot, LAS * 100, measures.sUAS, measures.stot, UAS * 100) + "%d / %d * 100 = %.2f%%\n---\n" % \
		(measures.sAcc, measures.stot, Acc * 100)).encode('utf-8')
	else:
		Prec = "NaN"
		output = ("\n  Precision (LAS): NaN\n  Precision (UAS): NaN\n  System accuracy: NaN\n---\n").encode('utf-8')
		
	if measures.gtot:
		LAS = float(measures.gLAS) / (measures.gtot)
		UAS = float(measures.gUAS) / (measures.gtot)
		Acc = float(measures.gAcc) / (measures.gtot)
		Rec = (LAS, UAS)
		output += ("  Recall (LAS):    %d / %d * 100 = %.2f%%\n  Recall (UAS):    %d / %d * 100 = %.2f%%\n  Gold accuracy:  " % \
		(measures.gLAS, measures.gtot, LAS * 100, measures.gUAS, measures.gtot, UAS * 100) + " %d / %d * 100 = %.2f%%\n---\n" % \
		(measures.gAcc, measures.gtot, Acc * 100)).encode('utf-8')
	else:
		Rec = "NaN"
		output += ("  Recall (LAS):    NaN\n  Recall (UAS):    NaN\n  Gold accuracy:   NaN\n---\n").encode('utf-8')
	
	if isinstance(Rec, str) or isinstance(Prec, str):
		output += ("  F1 measure (LAS): NaN\n  F1 measure (UAS): NaN\n").encode('utf-8')
	else:
		output += ("  F1 measure (LAS): %.2f%%\n" % ((2 * (Prec[0] * Rec[0]) / (Prec[0] + Rec[0])) * 100) if Prec[0] + Rec[0] else "  F1 measure (LAS): NaN\n").encode('utf-8')
		output += ("  F1 measure (UAS): %.2f%%\n\n" % ((2 * (Prec[1] * Rec[1]) / (Prec[1] + Rec[1])) * 100) if Prec[1] + Rec[1] else "  F1 measure (UAS): NaN\n\n").encode('utf-8')
			
	return output
	


# OUTPUT ADVANCED:
# calcola i valori delle prepc* e delle prep* creando due misure aggregate
def aggregatePrep():	
	prepc = [0, 0, 0, 0]
	prep = [0, 0, 0, 0]
	
	for k in reversed(sorted(all_measures)):
		if k.find("prepc") == 0:
			for i in range(0, 4):
				prepc[i] += all_measures[k][i]
				
		elif k.find("prep") == 0:
			for i in range(0, 4):
				prep[i] += all_measures[k][i]
				
	all_measures["prepc*"] = prepc
	all_measures["prep*"] = prep
			
			
			
# OUTPUT ADVANCED:
# calcola la precision, recall e f1 di tutte le dipendenze e restituisce una stringa con il messaggio
def getStringMeasures():
	output = ("================================================================================\n\n" + \
	"* Precision and recall of DEPREL\n" + \
	"  ----------------------+--------+---------+----------+------------+---------------\n" + \
	"  deprel                |  gold  | correct |  system  | recall (%) | precision (%) \n" + \
	"  ----------------------+--------+---------+----------+------------+---------------\n").encode('utf-8')
	# [recall: correct/gold] [precision: correct/system]
	for k in sorted(all_measures):
		rec = "%.2f" % ((float(all_measures[k][2]) / all_measures[k][0]) * 100) if all_measures[k][0] else "NaN"
		prec = "%.2f" % ((float(all_measures[k][2]) / all_measures[k][1]) * 100) if all_measures[k][1] else "NaN"
			
		output += ("  %s | %s | %s | %s | %s | %s \n" % (k.ljust(21), "{0:-6}".format(all_measures[k][0]), "{0:-7}".format(all_measures[k][2]), \
		"{0:-8}".format(all_measures[k][1]), rec.rjust(10), prec.rjust(13))).encode('utf-8')
	
	output += ("\n\n* Precision and recall of DEPREL + ATTACHMENT\n" + \
	"  ----------------------+--------+---------+----------+------------+---------------\n" + \
	"  deprel                |  gold  | correct |  system  | recall (%) | precision (%) \n" + \
	"  ----------------------+--------+---------+----------+------------+---------------\n").encode('utf-8')
	# [recall: correct/gold] [precision: correct/system]
	for k in sorted(all_measures):
		rec = "%.2f" % ((float(all_measures[k][3]) / all_measures[k][0]) * 100) if all_measures[k][0] else "NaN"
		prec = "%.2f" % ((float(all_measures[k][3]) / all_measures[k][1]) * 100) if all_measures[k][1] else "NaN"
			
		output += ("  %s | %s | %s | %s | %s | %s \n" % (k.ljust(21), "{0:-6}".format(all_measures[k][0]), "{0:-7}".format(all_measures[k][3]), \
		"{0:-8}".format(all_measures[k][1]), rec.rjust(10), prec.rjust(13))).encode('utf-8')
	
	output += ("\n").encode('utf-8')
	return output
			
			

# restituisce True se word è formata di soli caratteri di punteggiatura
def isPunctuation(word):
	ok = True
	for c in word:
		if not unicodedata.category(unichr(ord(c))).startswith('P'):
			ok = False
			break
	
	return ok
	
	
		
# trova le righe che sono true positive in SYS considerando il GOLD
# DATO IL TOKEN DIPENDENTE, calcola le seguenti misure:
# LAS: matchHEAD + matchDEP : O--O
# UAS: matchHEAD : O  O
# Acc: matchDEP	: --O	
# all_measures: 0 - in gold; 1 - in sys; 2 - Acc; 3 - LAS
def calcMeasures(GOLD, SYS, punct):
	measure = Measures()

  # for each line in sys, find the corrispective in gold (precision)
	for lines in SYS.rows:
		if not punct and isPunctuation(lines.body) or lines.dep in exclude_dep:
			continue
			
		measure.stot += 1
		# aggiorna o crea la misura relativa alla dipendenza in esame
		if all_measures.has_key(lines.dep):
			value = all_measures[lines.dep]
			value[1] += 1 # ha trovato la relazione nel file sys
		else:
			value = [0, 1, 0, 0] # è la prima del file sys
		
		lists = []	
		for lineg in GOLD.rows:				
			if lines.matchBODY(lineg):
				# print str(lines) + " - " + str(lineg)
				
				# se è presente un match LAS, ok va bene e termina la ricerca
				# perché non ci possono essere due dipendenze uguali con uguale
				# testa e uguale dipendente.
				if lines.matchDEP(lineg) and lines.matchHEAD(lineg):
					measure.sLAS += 1
					lines.info = "LAS"
					value[2] += 1 # ACC incluso in LAS
					value[3] += 1 
					break
					
				# altrimenti è UAS o LAS: memorizza tutte le dipendenze trovate e:
				# - se è stata trovata una UAS, la segna come UAS; altrimenti come Acc
				elif lines.matchHEAD(lineg):
					lists += ["UAS"] 
				elif lines.matchDEP(lineg):
					lists += ["Acc"]
					
		# alla fine se c'è almeno una UAS, assegna UAS, altrimenti LAS
		if "UAS" in lists:
			measure.sUAS += 1
			lines.info = "UAS"
		elif "Acc" in lists:
			measure.sAcc += 1
			value[2] += 1
			lines.info = "Acc"
		
		# sovrascrive l'aggiornamento
		all_measures[lines.dep] = value

	measure.sUAS += measure.sLAS
	measure.sAcc += measure.sLAS
  
	# for each line in gold, find the corrispective in sys (recall)
	for lineg in GOLD.rows:
		if not punct and isPunctuation(lineg.body) or lineg.dep in exclude_dep:
			continue
			
		measure.gtot += 1
		if all_measures.has_key(lineg.dep):
			value = all_measures[lineg.dep]
			value[0] += 1 # ha trovato la relazione nel file gold
			all_measures[lineg.dep] = value
		else:
			all_measures[lineg.dep] = [1, 0, 0, 0] # è la prima del file gold
		
		listg = []	
		for lines in SYS.rows:				
			if lineg.matchBODY(lines):
				# se è presente un match LAS, ok va bene e termina la ricerca
				# perché non ci possono essere due dipendenze uguali con uguale
				# testa e uguale dipendente.
				if lineg.matchDEP(lines) and lines.matchHEAD(lines):
					measure.gLAS += 1
					lineg.info = "LAS"
					break
				elif lineg.matchHEAD(lines):
					listg += ["UAS"] 
				elif lineg.matchDEP(lines):
					listg += ["Acc"]

		# alla fine se c'è almeno una UAS, assegna UAS, altrimenti LAS
		if "UAS" in listg:
			measure.gUAS += 1
			lines.info = "UAS"
		elif "Acc" in listg:
			measure.gAcc += 1
			lines.info = "Acc"
	
	measure.gUAS += measure.gLAS
	measure.gAcc += measure.gLAS
	
	return (measure)
			
	
			
def main():
	outfile = goldfile = systemfile = ""
	punct = False # default: not to score
	numOpts = 0
	measures = Measures()
	global diff_measures
	
	# global ferr
	# ferr = codecs.open(errorFileName, 'w')
	
	try:
		opts, args = getopt.getopt(sys.argv[1:],"dg:ho:s:")
	except getopt.GetoptError:
		exit("Error in input options or arguments.\n\nFor info, see manual:\n", True)
	
	for opt, value in opts: # opts è una lista di coppie (opzione, valore)
		if opt == '-h':
			print __doc__
			sys.exit()
		elif opt == '-g':
			goldfile = value
			numOpts += 1
		elif opt == '-d':
			diff_measures = True
		elif opt == '-o':
			outfile = value
		elif opt == '-s':
			systemfile = value
			numOpts += 1
		else:
			exit("Error in input options or arguments.\n\nFor info, see manual:\n", True)
	
	if numOpts > 2:
		exit("Error: too many options.\n\nFor info, see manual:\n", True)
	elif numOpts < 2:
		exit("Error: not enough options.\n\nFor info, see manual:\n", True)
		
	if len(args) != 0:
		exit("Error: too many arguments.\n\nFor info, see manual:", True)
	
	try:
		fgold = codecs.open(goldfile, 'r', 'utf-8')
		fsys = codecs.open(systemfile, 'r', 'utf-8')
	except:
		exit("Error while opening input files.\n", False)
		
	fout = codecs.open(outfile, 'w') if outfile else None
	# ferrGOLD = codecs.open(errorFileNameG, 'w')
	# ferrSYS = codecs.open(errorFileNameS, 'w')
	
	nlineG = nlineS = 1
	rows = []
	
	# legge le righe in fgold...
	for line1 in fgold:
		l = line1.strip()
		
		# fin quando non trova una riga bianca...
		if l:					
			# crea una nuova riga con le informazioni rilevanti lette da file e la inserisce nella lista rows
			rows.append(Row(l, nlineG))
			
		# non appena trova una riga bianca...
		else:
			# crea la struttura GOLD
			GOLD = Sentence(rows)
			
			rows = []
			# inizia a cercare la rispettiva frase nell'altro file, leggendo le righe fino ad incontrare
			# una riga vuota
			for line2 in fsys:
				l = line2.strip()
				
				if l:
					# crea una nuova riga con le informazioni rilevanti lette da file e la inserisce nella lista rows
					rows.append(Row(l, nlineS))
					nlineS += 1
				else:
					break
					
			# appena incontra una riga vuota, crea la struttura SYS ed inizia il confronto
			nlineS += 1
			SYS = Sentence(rows)	
			rows = []
			
			# print "match ", GOLD.rows[0], " with ", SYS.rows[0]
			# print findTP(GOLD, SYS, 'l'), SYS.getFALSE(), GOLD.getFALSE()
			# print findTP(GOLD, SYS, 'u'), SYS.getFALSE(), GOLD.getFALSE()
			
			# calcola le misure a aggiorna il conto complessivo
			measures.update(calcMeasures(GOLD, SYS, punct))
		
		nlineG += 1
		
		
	# nel caso in cui il file fgold non termina con la riga vuota:	
	if len(rows) > 0: 
		# crea la struttura GOLD
		GOLD = Sentence(rows)
		
		rows = []
		# inizia a cercare la rispettiva frase nell'altro file, leggendo le righe fino ad incontrare
		# una riga vuota
		for line2 in fsys:
			l = line2.strip()
			
			if l:
				# crea una nuova riga con le informazioni rilevanti lette da file e la inserisce nella lista rows
				rows.append(Row(l, nlineS))
				nlineS += 1
			else:
				break
			
		# appena incontra una riga vuota, crea la struttura SYS ed inizia il confronto
		nlineS += 1
		SYS = Sentence(rows)
		
		# calcola le misure a aggiorna il conto complessivo
		measures.update(calcMeasures(GOLD, SYS, punct))			
	
	# risultati ed eventuali dettagli
	res = getStringResult(measures)
	if diff_measures:
			aggregatePrep()
			res += getStringMeasures()
	
	# ferrGOLD.close()	
	# ferrSYS.close()
	fgold.close()
	fsys.close()
	if outfile:
		fout.write(res)
		fout.close()
	else:
		sys.stdout.write(res)
	
	# if not ferr.tell():
	#		ferr.write("No error was found.\n")
	# ferr.close()


if __name__ == '__main__':
  main()
